import os
import sys
import json
import threading
import subprocess
import time
import ctypes
from ctypes import pythonapi as cpyapi

from .. import _state
from ..vendor import six
from ..vendor.Qt import QtWidgets, QtCore

CREATE_NO_WINDOW = 0x08000000
IS_WIN32 = sys.platform == "win32"


SIGNALS_TO_CLOSE_VESSEL = (
    "pyblishQmlClose",
    "pyblishQmlCloseForced",
)


def default_wrapper(func, *args, **kwargs):
    return func(*args, **kwargs)


class FosterVessel(QtWidgets.QDialog):
    """Container window for pyblish-qml App in child process
    """

    def __init__(self, proxy):
        super(FosterVessel, self).__init__(_state.get("vesselParent"))

        # (NOTE) Although minimizing this vesselized dialog works well on
        #   Windows 10, but QQuickView window did not hide well on Windows 7.
        #   Other platform not tested yet, disable it for now.
        #
        # Enable maximize for app, minimize disabled
        self.setWindowFlags(QtCore.Qt.Window |
                            QtCore.Qt.WindowCloseButtonHint |
                            QtCore.Qt.WindowMaximizeButtonHint)

        self._winId = winIdFixed(self.winId())

        self.resize(1, 1)

        self.proxy = proxy
        self.close_lock = True

    def closeEvent(self, event):
        if self.close_lock:
            self.proxy.quit()
            event.ignore()
        else:
            event.accept()

    def resizeEvent(self, event):
        self.proxy._dispatch("resize", args=[self.width(), self.height()])


class MockFosterVessel(object):
    """We don't create widget without QApp, we mock one"""
    _winId = close_lock = None
    show = hide = close = lambda _: None


class Proxy(object):
    """Speak to child process and control the vessel (window container)"""

    def __init__(self, server):
        import pyblish.api

        self.popen = server.popen
        self.foster = server.foster
        self.foster_fixed = server.foster_fixed

        self.vessel = FosterVessel(self) if self.foster else MockFosterVessel()
        self._winId = self.vessel._winId

        server.proxy = self

        def close_vessel():
            self.vessel.close_lock = False
            self.vessel.close()

            for signal in SIGNALS_TO_CLOSE_VESSEL:
                try:
                    pyblish.api.deregister_callback(signal, self.close_vessel)
                except (KeyError, ValueError):
                    pass

        self.close_vessel = close_vessel

        for signal in SIGNALS_TO_CLOSE_VESSEL:
            pyblish.api.register_callback(signal, self.close_vessel)

        self._alive()

    def show(self, settings=None):
        """Show the GUI

        Arguments:
            settings (optional, dict): Client settings

        """
        if self.foster_fixed:
            self.vessel.show()
        return self._dispatch("show", args=[settings or {},
                                            self._winId,
                                            self.foster_fixed])

    def hide(self):
        """Hide the GUI"""
        success = self._dispatch("hide")
        self.vessel.hide()
        return success

    def quit(self):
        """Ask the GUI to quit"""
        self._dispatch("quit")

    def rise(self):
        """Rise GUI from hidden"""
        self.vessel.show()
        return self._dispatch("rise")

    def inFocus(self):
        """Set GUI on-top flag"""
        return self._dispatch("inFocus")

    def outFocus(self):
        """Remove GUI on-top flag"""
        return self._dispatch("outFocus")

    def kill(self):
        """Forcefully destroy the process"""
        self.popen.kill()

    def detach(self):
        self.vessel.setWindowOpacity(0)  # avoid hide window anim
        self.vessel.hide()
        self._dispatch("host_detach")

    def attach(self, x, y, w, h):
        self.vessel.show()
        self.vessel.setGeometry(x, y, w, h)
        self.vessel.setWindowOpacity(1)
        self._dispatch("host_attach")

    def popup(self, alert):
        # No hijack keyboard focus
        if not self.foster_fixed:
            QtWidgets.QApplication.setActiveWindow(self.vessel)
        # Plus alert
        if alert:
            QtWidgets.QApplication.alert(self.vessel.parent(), 0)

    def publish(self):
        return self._dispatch("publish")

    def validate(self):
        return self._dispatch("validate")

    def _alive(self):
        """Send pulse to child process

        Child process will run forever if parent process encounter such
        failure that not able to kill child process.

        This inform child process that server is still running and child
        process will auto kill itself after server stop sending pulse
        message.

        """

        def _pulse():
            start_time = time.time()

            while True:
                data = json.dumps({"header": "pyblish-qml:server.pulse"})
                if not self._flush(data):
                    break

                # Send pulse every 5 seconds
                time.sleep(5.0 - ((time.time() - start_time) % 5.0))

        thread = threading.Thread(target=_pulse)
        thread.daemon = True
        thread.start()

    def _dispatch(self, func, args=None):
        data = json.dumps(
            {
                "header": "pyblish-qml:popen.parent",
                "payload": {
                    "name": func,
                    "args": args or list(),
                }
            }
        )

        return self._flush(data)

    def _flush(self, data):
        if six.PY3:
            data = data.encode("ascii")

        try:
            self.popen.stdin.write(data + b"\n")
            self.popen.stdin.flush()
        except IOError:
            # subprocess closed
            self.close_vessel()
        else:
            return True


class Server(object):
    """A subprocess server

    This server relies on stdout and stdin for interprocess communication.

    Arguments:
        service (service.Service): Dispatch requests to this service
        python (str, optional): Absolute path to Python executable
        pyqt5 (str, optional): Absolute path to PyQt5
        targets (list, optional): Publishing targets, e.g. `ftrack`
        modal (bool, optional): Block interactions to parent
        foster (bool, optional): GUI become a real child of the parent process
        foster_fixed (bool, optional): GUI always remain inside the parent

    """

    def __init__(self,
                 service,
                 python=None,
                 pyqt5=None,
                 targets=[],
                 modal=False,
                 foster=False,
                 foster_fixed=False):
        super(Server, self).__init__()
        self.service = service
        self.listening = False

        self.proxy = None

        # Store modal state
        self.modal = modal

        self.foster = foster
        self.foster_fixed = foster_fixed

        # The server may be run within Maya or some other host,
        # in which case we refer to it as running embedded.
        is_embedded = os.path.split(sys.executable)[-1].lower() != "python.exe"

        python = python or find_python()
        pyqt5 = pyqt5 or find_pyqt5(python)

        print("Using Python @ '%s'" % python)
        print("Using PyQt5 @ '%s'" % pyqt5)

        # Maintain the absolute minimum of environment variables,
        # to avoid issues on invalid types.
        if IS_WIN32:
            environ = {
                key: os.getenv(key)
                for key in ("USERNAME",
                            "SYSTEMROOT",
                            "PYTHONPATH",
                            "PATH")
                if os.getenv(key)
            }
        else:
            # OSX and Linux users are left to fend for themselves.
            environ = os.environ.copy()

        # Append PyQt5 to existing PYTHONPATH, if available
        environ["PYTHONPATH"] = os.pathsep.join(
            path for path in [os.getenv("PYTHONPATH"), pyqt5]
            if path is not None
        )

        # Protect against an erroneous parent environment
        # The environment passed to subprocess is inherited from
        # its parent, but the parent may - at run time - have
        # made the environment invalid. For example, a unicode
        # key or value in `os.environ` is not a valid environment.
        if not all(isinstance(key, str) for key in environ):
            print("One or more of your environment variable "
                  "keys are not <str>")

            for key in os.environ:
                if not isinstance(key, str):
                    print("- %s is not <str>" % key)

        if not all(isinstance(key, str) for key in environ.values()):
            print("One or more of your environment "
                  "variable values are not <str>")

            for key, value in os.environ.items():
                if not isinstance(value, str):
                    print("- Value of %s=%s is not <str>" % (key, value))

        kwargs = dict(args=[
            python, "-u", "-m", "pyblish_qml",

            # Indicate that this is a child of the parent process,
            # and that it should expect to speak with the parent
            "--aschild"],

            # Any path other than where the host was launched from
            # to prevent accidental pickup of e.g. PyQt5 binaries.
            cwd=os.path.dirname(__file__),

            env=environ,

            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

        if IS_WIN32 and is_embedded:
            # This will prevent an embedded Python
            # from opening an external terminal window.
            kwargs["creationflags"] = CREATE_NO_WINDOW

        # If no targets are passed to pyblish-qml, we assume that we want the
        # default target and the registered targets. This is to facilitate
        # getting all plugins on pyblish_qml.show().
        import pyblish.api
        if not targets:
            targets = ["default"] + pyblish.api.registered_targets()
        print("Targets: {0}".format(", ".join(targets)))

        kwargs["args"].append("--targets")
        kwargs["args"].extend(targets)

        self.popen = subprocess.Popen(**kwargs)

    def stop(self):
        try:
            return self.popen.kill()
        except OSError as e:
            # Assume process is already dead
            print(e)

    def wait(self):
        return self.popen.wait()

    def listen(self):
        """Listen to both stdout and stderr

        We'll want messages of a particular origin and format to
        cause QML to perform some action. Other messages are simply
        forwarded, as they are expected to be plain print or error messages.

        """

        def _listen():
            """This runs in a thread"""
            for line in iter(self.popen.stdout.readline, b""):

                if six.PY3:
                    line = line.decode("utf8")

                try:
                    response = json.loads(line)

                except Exception:
                    # This must be a regular message.
                    sys.stdout.write(line)

                else:
                    if response.get("header") == "pyblish-qml:popen.request":
                        payload = response["payload"]
                        args = payload["args"]

                        func_name = payload["name"]

                        # self.service have no access to proxy object, so
                        # this `if` statement is needed
                        if func_name in ("detach", "attach", "popup"):
                            getattr(self.proxy, func_name)(*args)
                            result = None

                        elif func_name in ("emit",):
                            # Avoid main thread hang
                            result = getattr(self.service, func_name)(*args)

                        else:
                            wrapper = _state.get("dispatchWrapper",
                                                 default_wrapper)

                            func = getattr(self.service, func_name)
                            result = wrapper(func, *args)  # block..

                        # Note(marcus): This is where we wait for the host to
                        # finish. Technically, we could kill the GUI at this
                        # point which would make the following commands throw
                        # an exception. However, no host is capable of kill
                        # the GUI whilst running a command. The host is locked
                        # until finished, which means we are guaranteed to
                        # always respond.

                        data = json.dumps({
                            "header": "pyblish-qml:popen.response",
                            "payload": result
                        })

                        if six.PY3:
                            data = data.encode("ascii")

                        try:
                            self.popen.stdin.write(data + b"\n")
                            self.popen.stdin.flush()
                        except IOError:
                            # subprocess closed
                            return

                    else:
                        # In the off chance that a message
                        # was successfully decoded as JSON,
                        # but *wasn't* a request, just print it.
                        sys.stdout.write(line)

        if not self.listening:
            if self.modal:
                _listen()
            else:
                thread = threading.Thread(target=_listen)
                thread.daemon = True
                thread.start()

            self.listening = True


def find_python():
    """Search for Python automatically"""
    python = (
        _state.get("pythonExecutable") or

        # Support for multiple executables.
        next((
            exe for exe in
            os.getenv("PYBLISH_QML_PYTHON_EXECUTABLE", "").split(os.pathsep)
            if os.path.isfile(exe)), None
        ) or

        # Search PATH for executables.
        which("python") or
        which("python3")
    )

    if not python or not os.path.isfile(python):
        raise ValueError("Could not locate Python executable.")

    return python


def find_pyqt5(python):
    """Search for PyQt5 automatically"""
    pyqt5 = (
        _state.get("pyqt5") or
        os.getenv("PYBLISH_QML_PYQT5")
    )

    return pyqt5


def which(program):
    """Locate `program` in PATH

    Arguments:
        program (str): Name of program, e.g. "python"

    """

    def is_exe(fpath):
        if os.path.isfile(fpath) and os.access(fpath, os.X_OK):
            return True
        return False

    for path in os.environ["PATH"].split(os.pathsep):
        for ext in os.getenv("PATHEXT", "").split(os.pathsep):
            fname = program + ext.lower()
            abspath = os.path.join(path.strip('"'), fname)

            if is_exe(abspath):
                return abspath

    return None


def winIdFixed(winId):
    # PySide bug: QWidget.winId() returns <PyCObject object at 0x02FD8788>,
    # there is no easy way to convert it to int.
    try:
        return int(winId)
    except Exception:
        if sys.version_info[0] == 2:
            cpyapi.PyCObject_AsVoidPtr.restype = ctypes.c_void_p
            cpyapi.PyCObject_AsVoidPtr.argtypes = [ctypes.py_object]

            return cpyapi.PyCObject_AsVoidPtr(winId)

        elif sys.version_info[0] == 3:
            cpyapi.PyCapsule_GetPointer.restype = ctypes.c_void_p
            cpyapi.PyCapsule_GetPointer.argtypes = [ctypes.py_object]

            return cpyapi.PyCapsule_GetPointer(winId, None)



import os
import sys
import inspect
import traceback

import pyblish.api

from . import ipc, settings, _state
from .vendor.Qt import QtWidgets, QtCore, QtGui

MODULE_DIR = os.path.dirname(__file__)
SPLASH_PATH = os.path.join(MODULE_DIR, "splash.png")


def register_dispatch_wrapper(wrapper):
    """Register a dispatch wrapper for servers

    The wrapper must have this exact signature:
        (func, *args, **kwargs)

    """

    signature = inspect.getargspec(wrapper)
    if any([len(signature.args) != 1,
            signature.varargs is None,
            signature.keywords is None]):
        raise TypeError("Wrapper signature mismatch")

    def _wrapper(func, *args, **kwargs):
        """Exception handling"""
        try:
            return wrapper(func, *args, **kwargs)
        except Exception as e:
            # Kill subprocess
            _state["currentServer"].stop()
            traceback.print_exc()
            raise e

    _state["dispatchWrapper"] = _wrapper


def deregister_dispatch_wrapper():
    _state.pop("dispatchWrapper")


def dispatch_wrapper():
    return _state.get("dispatchWrapper")


def current_server():
    return _state.get("currentServer")


def install(modal, foster):
    """Perform first time install

    Attributes:
        initial_port (int, optional): Port from which to start
            looking for available ports, defaults to 9001

    """

    if _state.get("installed"):
        sys.stdout.write("Already installed, uninstalling..\n")
        uninstall()

    use_threaded_wrapper = not modal

    install_callbacks()
    install_host(use_threaded_wrapper)

    _state["installed"] = True


def uninstall():
    """Clean up traces of Pyblish QML"""
    uninstall_callbacks()
    sys.stdout.write("Pyblish QML shutdown successful.\n")


def _is_headless():
    app = QtWidgets.QApplication.instance()
    return (
        # Maya 2017+ in standalone
        not hasattr(app, "activeWindow") or

        # Maya 2016-
        not app.activeWindow()
    )


def _fosterable(foster, modal):
    if foster is None:
        # Get foster mode from environment
        foster = bool(os.environ.get("PYBLISH_QML_FOSTER", False))

    if foster:
        if modal or _is_headless():
            print("Foster disabled due to Modal is on or in headless mode.")
            return False

        print("Foster on.")
        return True

    else:
        return False


def _foster_fixed(foster):
    if not foster:
        return False

    value = os.environ.get("PYBLISH_QML_FOSTER_FIXED", "").lower()
    return value in ("true", "yes", "1") or QtCore.qVersion()[0] == "4"


def show(parent=None, targets=[], modal=None, foster=None):
    """Attempt to show GUI

    Requires install() to have been run first, and
    a live instance of Pyblish QML in the background.

    Arguments:
        parent (None, optional): Deprecated
        targets (list, optional): Publishing targets
        modal (bool, optional): Block interactions to parent
        foster (bool, optional): Become a real child of the parent process

    """

    # Get modal mode from environment
    if modal is None:
        modal = bool(os.environ.get("PYBLISH_QML_MODAL", False))

    is_headless = _is_headless()

    foster = _fosterable(foster, modal)
    foster_fixed = _foster_fixed(foster)

    # Automatically install if not already installed.
    if not _state.get("installed"):
        install(modal, foster)

    # Show existing GUI
    if _state.get("currentServer"):
        server = _state["currentServer"]
        proxy = server.proxy or ipc.server.Proxy(server)

        if proxy.show(settings.to_dict()):
            return server

        else:
            # The running instance has already been closed.
            _state.pop("currentServer")

    if not is_headless:
        # mayapy would have a QtGui.QGuiApplication
        splash = Splash()
        splash.show()

        def on_shown():
            try:
                splash.close()

            except RuntimeError:
                # Splash already closed
                pass

            pyblish.api.deregister_callback(*callback)

        callback = "pyblishQmlShown", on_shown
        pyblish.api.register_callback(*callback)
    else:
        def on_shown():
            pass

    try:
        service = ipc.service.Service()
        server = ipc.server.Server(service,
                                   targets=targets,
                                   modal=modal,
                                   foster=foster,
                                   foster_fixed=foster_fixed)
    except Exception:
        # If for some reason, the GUI fails to show.
        traceback.print_exc()
        return on_shown()

    proxy = ipc.server.Proxy(server)
    proxy.show(settings.to_dict())

    # Store reference to server for future calls
    _state["currentServer"] = server

    print("Success. QML server available as "
          "pyblish_qml.api.current_server()")

    # Install eventFilter if not exists one
    install_event_filter()

    server.listen()

    return server


def publish():
    # get existing GUI
    if _state.get("currentServer"):
        server = _state["currentServer"]
        proxy = server.proxy or ipc.server.Proxy(server)

        if not proxy.publish():
            # The running instance has already been closed.
            _state.pop("currentServer")


def validate():
    # get existing GUI
    if _state.get("currentServer"):
        server = _state["currentServer"]
        proxy = server.proxy or ipc.server.Proxy(server)

        if not proxy.validate():
            # The running instance has already been closed.
            _state.pop("currentServer")


def install_callbacks():
    pyblish.api.register_callback("instanceToggled", _toggle_instance)
    pyblish.api.register_callback("pluginToggled", _toggle_plugin)


def uninstall_callbacks():
    pyblish.api.deregister_callback("instanceToggled", _toggle_instance)
    pyblish.api.deregister_callback("pluginToggled", _toggle_plugin)


def _toggle_instance(instance, new_value, old_value):
    """Alter instance upon visually toggling it"""
    instance.data["publish"] = new_value


def _toggle_plugin(plugin, new_value, old_value):
    """Alter plugin upon visually toggling it"""
    plugin.active = new_value


def register_python_executable(path):
    """Expose Python executable to server

    The Python executable must be compatible with the
    version of PyQt5 installed or provided on the system.

    """

    assert os.path.isfile(path), "Must be a file, such as python.exe"

    _state["pythonExecutable"] = path


def registered_python_executable():
    return _state.get("pythonExecutable")


def register_pyqt5(path):
    """Expose PyQt5 to Python

    The exposed PyQt5 must be compatible with the exposed Python.

    Arguments:
        path (str): Absolute path to directory containing PyQt5

    """

    _state["pyqt5"] = path


def install_host(use_threaded_wrapper):
    """Install required components into supported hosts

    An unsupported host will still run, but may encounter issues,
    especially with threading.

    """

    for install in (_install_maya,
                    _install_houdini,
                    _install_nuke,
                    _install_nukeassist,
                    _install_hiero,
                    _install_nukestudio):
        try:
            install(use_threaded_wrapper)
        except ImportError:
            pass
        else:
            break


SIGNALS_TO_REMOVE_EVENT_FILTER = (
    "pyblishQmlClose",
    "pyblishQmlCloseForced",
)


def remove_event_filter():
    event_filter = _state.get("eventFilter")
    if isinstance(event_filter, QtCore.QObject):

        # (NOTE) Should remove from the QApp instance which originally
        #        installed to.
        #        This will not work:
        #        `QApplication.instance().removeEventFilter(event_filter)`
        #
        event_filter.parent().removeEventFilter(event_filter)
        del _state["eventFilter"]

        for signal in SIGNALS_TO_REMOVE_EVENT_FILTER:
            try:
                pyblish.api.deregister_callback(signal, remove_event_filter)
            except (KeyError, ValueError):
                pass

        print("The eventFilter of pyblish-qml has been removed.")


def install_event_filter():

    main_window = _state.get("vesselParent")
    if main_window is None:
        print("Main window not found, event filter did not install.")
        return

    event_filter = _state.get("eventFilter")
    if isinstance(event_filter, QtCore.QObject):
        print("Event filter exists.")
        return
    else:
        event_filter = HostEventFilter(main_window)

    try:
        main_window.installEventFilter(event_filter)
    except Exception as e:
        print("An error has occurred during event filter's installation.")
        print(e)
    else:
        _state["eventFilter"] = event_filter

        for signal in SIGNALS_TO_REMOVE_EVENT_FILTER:
            pyblish.api.register_callback(signal, remove_event_filter)

        print("Event filter has been installed.")


def _on_application_quit():
    """Automatically kill QML on host exit"""

    try:
        _state["currentServer"].popen.kill()

    except KeyError:
        # No server started
        pass

    except OSError:
        # Already dead
        pass


class HostEventFilter(QtWidgets.QWidget):
    """Connect some event from host to QML

    Host will connect following event to QML:
    QEvent.Show                -> rise QML
    QEvent.Hide                -> hide QML
    QEvent.WindowActivate      -> set QML on top
    QEvent.WindowDeactivate    -> remove QML on top
    """

    eventList = {
        QtCore.QEvent.Show: "rise",
        QtCore.QEvent.Hide: "hide",
        QtCore.QEvent.WindowActivate: "inFocus",
        QtCore.QEvent.WindowDeactivate: "outFocus",
    }

    def eventFilter(self, widget, event):

        try:
            func_name = self.eventList[event.type()]
        except KeyError:
            return False

        server = _state.get("currentServer")
        proxy = server.proxy if server else None

        try:
            func = getattr(proxy, func_name)
        except AttributeError:
            # proxy is None, or does not have the function
            return False

        connected = func()

        if connected is not True:
            # The running instance has already been closed.
            remove_event_filter()

        return True


def _acquire_host_main_window(app):

    # Get top window in host
    _window = app.activeWindow()
    while True:
        parent_window = _window.parent()
        if parent_window:
            _window = parent_window
        else:
            break

    _state["vesselParent"] = _window


def _set_host_label(host_name):
    if settings.ContextLabel == settings.ContextLabelDefault:
        settings.ContextLabel = host_name
    if settings.WindowTitle == settings.WindowTitleDefault:
        settings.WindowTitle = "Pyblish ({})".format(host_name)


def _remove_googleapiclient():
    """Check if the compatibility must be maintained

    The Maya 2018 version tries to import the `http` module from
    Maya2018\plug-ins\MASH\scripts\googleapiclient\http.py in stead of the
    module from six.py. This import conflict causes a crash Avalon's publisher.
    This is due to Autodesk adding paths to the PYTHONPATH environment variable
    which contain modules instead of only packages.
    """

    keyword = "googleapiclient"

    # reconstruct python paths
    python_paths = os.environ["PYTHONPATH"].split(os.pathsep)
    paths = [path for path in python_paths if keyword not in path]
    os.environ["PYTHONPATH"] = os.pathsep.join(paths)


def _install_maya(use_threaded_wrapper):
    """Helper function to Autodesk Maya support"""
    from maya import utils, cmds

    def threaded_wrapper(func, *args, **kwargs):
        return utils.executeInMainThreadWithResult(
            func, *args, **kwargs)

    sys.stdout.write("Setting up Pyblish QML in Maya\n")

    if use_threaded_wrapper:
        register_dispatch_wrapper(threaded_wrapper)

    if cmds.about(version=True) == "2018":
        _remove_googleapiclient()

    app = QtWidgets.QApplication.instance()

    if not _is_headless():
        # mayapy would have a QtGui.QGuiApplication
        app.aboutToQuit.connect(_on_application_quit)

        # acquire Maya's main window
        _state["vesselParent"] = {
            widget.objectName(): widget
            for widget in QtWidgets.QApplication.topLevelWidgets()
        }["MayaWindow"]

    _set_host_label("Maya")


def _common_setup(host_name, threaded_wrapper, use_threaded_wrapper):

    sys.stdout.write("Setting up Pyblish QML in {}\n".format(host_name))

    if use_threaded_wrapper:
        register_dispatch_wrapper(threaded_wrapper)

    app = QtWidgets.QApplication.instance()
    app.aboutToQuit.connect(_on_application_quit)
    _acquire_host_main_window(app)

    _set_host_label(host_name)


def _install_houdini(use_threaded_wrapper):
    """Helper function to SideFx Houdini support"""
    import hdefereval

    def threaded_wrapper(func, *args, **kwargs):
        return hdefereval.executeInMainThreadWithResult(
            func, *args, **kwargs)

    _common_setup("Houdini", threaded_wrapper, use_threaded_wrapper)


def _install_nuke(use_threaded_wrapper):
    """Helper function to The Foundry Nuke support"""
    import nuke

    not_nuke_launch = (
        "--hiero" in nuke.rawArgs or
        "--studio" in nuke.rawArgs or
        "--nukeassist" in nuke.rawArgs
    )
    if not_nuke_launch:
        raise ImportError

    def threaded_wrapper(func, *args, **kwargs):
        return nuke.executeInMainThreadWithResult(
            func, args, kwargs)

    _common_setup("Nuke", threaded_wrapper, use_threaded_wrapper)


def _install_nukeassist(use_threaded_wrapper):
    """Helper function to The Foundry NukeAssist support"""
    import nuke

    if "--nukeassist" not in nuke.rawArgs:
        raise ImportError

    def threaded_wrapper(func, *args, **kwargs):
        return nuke.executeInMainThreadWithResult(
            func, args, kwargs)

    _common_setup("NukeAssist", threaded_wrapper, use_threaded_wrapper)


def _install_hiero(use_threaded_wrapper):
    """Helper function to The Foundry Hiero support"""
    import hiero
    import nuke

    if "--hiero" not in nuke.rawArgs:
        raise ImportError

    def threaded_wrapper(func, *args, **kwargs):
        return hiero.core.executeInMainThreadWithResult(
            func, args, kwargs)

    _common_setup("Hiero", threaded_wrapper, use_threaded_wrapper)


def _install_nukestudio(use_threaded_wrapper):
    """Helper function to The Foundry Hiero support"""
    import nuke

    if "--studio" not in nuke.rawArgs:
        raise ImportError

    def threaded_wrapper(func, *args, **kwargs):
        return nuke.executeInMainThreadWithResult(
            func, args, kwargs)

    _common_setup("NukeStudio", threaded_wrapper, use_threaded_wrapper)


class Splash(QtWidgets.QWidget):
    """Splash screen for loading QML via subprocess

    Loading pyblish-qml may take some time, so when loading
    from within an existing interpreter, such as Maya, this
    splash screen can keep the user company during that time.

    """

    def __init__(self, parent=None):
        super(Splash, self).__init__(parent)
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.setAttribute(QtCore.Qt.WA_TranslucentBackground)
        self.setWindowFlags(
            QtCore.Qt.WindowStaysOnTopHint |
            QtCore.Qt.FramelessWindowHint
        )

        pixmap = QtGui.QPixmap(SPLASH_PATH)
        image = QtWidgets.QLabel()
        image.setPixmap(pixmap)

        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(image)

        label = QtWidgets.QLabel(self)
        label.move(20, 170)
        label.show()

        self.count = 0
        self.label = label

        self.setStyleSheet("""
            QLabel {
                color: white
            }
        """)

        loop = QtCore.QTimer()
        loop.timeout.connect(self.animate)
        loop.start(330)

        self.loop = loop

        self.animate()
        self.resize(200, 200)

    def animate(self):
        self.label.setText("loading" + "." * self.count)
        self.count = (self.count + 1) % 4



"""Application entry-point"""

# Standard library
import os
import sys
import time
import json
import traceback
import threading

# Dependencies
from PyQt5 import QtCore, QtGui, QtQuick, QtTest

# Local libraries
from . import util, compat, control, settings, ipc

MODULE_DIR = os.path.dirname(__file__)
QML_IMPORT_DIR = os.path.join(MODULE_DIR, "qml")
APP_PATH = os.path.join(MODULE_DIR, "qml", "main.qml")
ICON_PATH = os.path.join(MODULE_DIR, "icon.ico")


class Window(QtQuick.QQuickView):
    """Main application window"""

    def __init__(self):
        super(Window, self).__init__(None)
        self.app = QtGui.QGuiApplication.instance()

        self.setTitle(settings.WindowTitle)
        self.setResizeMode(self.SizeRootObjectToView)

        self.resize(*settings.WindowSize)
        self.setMinimumSize(QtCore.QSize(430, 300))

    def event(self, event):
        """Allow GUI to be closed upon holding Shift"""
        if event.type() == QtCore.QEvent.Close:
            modifiers = self.app.queryKeyboardModifiers()
            shift_pressed = QtCore.Qt.ShiftModifier & modifiers
            states = self.app.controller.states

            if shift_pressed:
                print("Force quitted..")
                self.app.controller.host.emit("pyblishQmlCloseForced")
                event.accept()

            elif any(state in states for state in ("ready", "finished")):
                self.app.controller.host.emit("pyblishQmlClose")
                event.accept()

            else:
                print("Not ready, hold SHIFT to force an exit")
                event.ignore()

        return super(Window, self).event(event)


class NativeVessel(QtGui.QWindow):
    """Container window"""

    def __init__(self):
        super(NativeVessel, self).__init__(None)
        self.app = QtGui.QGuiApplication.instance()

    def resizeEvent(self, event):
        self.app.resize(self.width(), self.height())

    def event(self, event):
        # Is required when Foster mode is on.
        # Native vessel will receive closeEvent while foster mode is on
        # and is the parent of window.
        if event.type() == QtCore.QEvent.Close:
            self.app.window.event(event)
            if event.isAccepted():
                # `app.fostered` is False at this moment.
                self.app.quit()

        return super(NativeVessel, self).event(event)


class Application(QtGui.QGuiApplication):
    """Pyblish QML wrapper around QGuiApplication

    Provides production and debug launchers along with controller
    initialisation and orchestration.

    """

    shown = QtCore.pyqtSignal(*(QtCore.QVariant,) * 3)
    hidden = QtCore.pyqtSignal()
    quitted = QtCore.pyqtSignal()
    published = QtCore.pyqtSignal()
    validated = QtCore.pyqtSignal()

    resized = QtCore.pyqtSignal(QtCore.QVariant, QtCore.QVariant)

    risen = QtCore.pyqtSignal()
    inFocused = QtCore.pyqtSignal()
    outFocused = QtCore.pyqtSignal()

    attached = QtCore.pyqtSignal(QtCore.QVariant)
    detached = QtCore.pyqtSignal()
    host_attached = QtCore.pyqtSignal()
    host_detached = QtCore.pyqtSignal()

    def __init__(self, source, targets=[]):
        super(Application, self).__init__(sys.argv)

        self.setWindowIcon(QtGui.QIcon(ICON_PATH))

        native_vessel = NativeVessel()

        window = Window()
        window.statusChanged.connect(self.on_status_changed)

        engine = window.engine()
        engine.addImportPath(QML_IMPORT_DIR)

        host = ipc.client.Proxy()
        controller = control.Controller(host, targets=targets)
        controller.finished.connect(lambda: window.alert(0))

        context = engine.rootContext()
        context.setContextProperty("app", controller)

        self.fostered = False
        self.foster_fixed = False

        self.foster_vessel = None
        self.native_vessel = native_vessel

        self.window = window
        self.engine = engine
        self.controller = controller
        self.host = host
        self.clients = dict()
        self.current_client = None

        self.shown.connect(self.show)
        self.hidden.connect(self.hide)
        self.quitted.connect(self.quit)
        self.published.connect(self.publish)
        self.validated.connect(self.validate)

        self.resized.connect(self.resize)

        self.risen.connect(self.rise)
        self.inFocused.connect(self.inFocus)
        self.outFocused.connect(self.outFocus)

        self.attached.connect(self.attach)
        self.detached.connect(self.detach)

        window.setSource(QtCore.QUrl.fromLocalFile(source))

    def on_status_changed(self, status):
        if status == QtQuick.QQuickView.Error:
            self.quit()

    def register_client(self, port):
        self.current_client = port
        self.clients[port] = {
            "lastSeen": time.time()
        }

    def deregister_client(self, port):
        self.clients.pop(port)

    def quit(self):
        event = None
        if self.fostered:
            # Foster vessel's closeEvent will trigger "quit" which connected
            # to here.
            # Forward the event to window.
            event = QtCore.QEvent(QtCore.QEvent.Close)
            self.window.event(event)

        if event is None or event.isAccepted():
            super(Application, self).quit()

    @util.SlotSentinel()
    def show(self, client_settings=None, window_id=None, foster_fixed=False):
        """Display GUI

        Once the QML interface has been loaded, use this
        to display it.

        Arguments:
            port (int): Client asking to show GUI.
            client_settings (dict, optional): Visual settings, see settings.py

        """
        self.fostered = window_id is not None

        if self.fostered:
            print("Moving to container window ...")

            # Creates a local representation of a window created by another
            # process (Maya or other host).
            foster_vessel = QtGui.QWindow.fromWinId(window_id)

            if foster_vessel is None:
                raise RuntimeError("Container window not found, ID: {}\n."
                                   "This is a bug.".format(window_id))

            self.window.setParent(foster_vessel)
            self.foster_vessel = foster_vessel
            self.foster_fixed = foster_fixed

        if client_settings:
            # Apply client-side settings
            settings.from_dict(client_settings)

            def first_appearance_setup(vessel):
                vessel.setGeometry(client_settings["WindowPosition"][0],
                                   client_settings["WindowPosition"][1],
                                   client_settings["WindowSize"][0],
                                   client_settings["WindowSize"][1])
                vessel.setTitle(client_settings["WindowTitle"])

            first_appearance_setup(self.native_vessel)

            if self.fostered:
                if not self.foster_fixed:
                    # Return it back to native vessel for first run
                    self.window.setParent(self.native_vessel)
                first_appearance_setup(self.foster_vessel)

        message = list()
        message.append("Settings: ")
        for key, value in settings.to_dict().items():
            message.append("  %s = %s" % (key, value))

        print("\n".join(message))

        if self.fostered and not self.foster_fixed:
            self.native_vessel.show()

        self.window.requestActivate()
        self.window.showNormal()

        self._popup()

        # Give statemachine enough time to boot up
        if not any(state in self.controller.states
                   for state in ["ready", "finished"]):
            util.timer("ready")

            ready = QtTest.QSignalSpy(self.controller.ready)

            count = len(ready)
            ready.wait(1000)
            if len(ready) != count + 1:
                print("Warning: Could not enter ready state")

            util.timer_end("ready", "Awaited statemachine for %.2f ms")

        self.controller.show.emit()

        # Allow time for QML to initialise
        util.schedule(self.controller.reset, 500, channel="main")

    def hide(self):
        """Hide GUI

        Process remains active and may be shown
        via a call to `show()`

        """
        self.window.hide()

    def rise(self):
        """Rise GUI from hidden"""
        self.window.show()

    def inFocus(self):
        """Set GUI on-top flag"""
        if not self.fostered:
            previous_flags = self.window.flags()
            self.window.setFlags(previous_flags |
                                 QtCore.Qt.WindowStaysOnTopHint)

    def outFocus(self):
        """Remove GUI on-top flag"""
        if not self.fostered:
            previous_flags = self.window.flags()
            self.window.setFlags(previous_flags ^
                                 QtCore.Qt.WindowStaysOnTopHint)

    def resize(self, width, height):
        """Resize GUI with it's vessel (container window)
        """
        # (NOTE) Could not get it auto resize with container, this is a
        #   alternative
        self.window.resize(width, height)

    def _popup(self):
        if not self.fostered:
            window = self.window
            # Work-around for window appearing behind
            # other windows upon being shown once hidden.
            previous_flags = window.flags()
            window.setFlags(previous_flags | QtCore.Qt.WindowStaysOnTopHint)
            window.setFlags(previous_flags)

    def detach(self):
        """Detach QQuickView window from the host

        In foster mode, inorder to prevent window freeze when the host's
        main thread is busy, will detach the QML window from the container
        inside the host, and re-parent to the container which spawned by
        the subprocess. And attach it back to host when the heavy lifting
        is done.

        This is the part that detaching from host.

        """
        if self.foster_fixed or self.foster_vessel is None:
            self.controller.detached.emit()
            return

        print("Detach window from foster parent...")

        self.fostered = False
        self.window.setParent(self.native_vessel)

        # Show dst container
        self.native_vessel.show()
        self.native_vessel.setGeometry(self.foster_vessel.geometry())
        self.native_vessel.setOpacity(1)
        # Hide src container (will wait for host)
        host_detached = QtTest.QSignalSpy(self.host_detached)
        self.host.detach()
        host_detached.wait(300)
        # Stay on top
        self.window.requestActivate()
        self._popup()

        self.controller.detached.emit()

    def attach(self, alert=False):
        """Attach QQuickView window to the host

        In foster mode, inorder to prevent window freeze when the host's
        main thread is busy, will detach the QML window from the container
        inside the host, and re-parent to the container which spawned by
        the subprocess. And attach it back to host when the heavy lifting
        is done.

        This is the part that attaching back to host.

        """
        if self.foster_fixed or self.foster_vessel is None:
            self.controller.attached.emit()
            if self.foster_vessel is not None:
                self.host.popup(alert)  # Send alert
            return

        print("Attach window to foster parent...")

        self.fostered = True
        self.window.setParent(self.foster_vessel)

        # Show dst container (will wait for host)
        host_attached = QtTest.QSignalSpy(self.host_attached)
        self.host.attach(self.native_vessel.geometry())
        host_attached.wait(300)
        # Hide src container
        self.native_vessel.setOpacity(0)  # avoid hide window anim
        self.native_vessel.hide()
        # Stay on top
        self.host.popup(alert)

        self.controller.attached.emit()

    def publish(self):
        """Fire up the publish sequence"""
        self.controller.publish()

    def validate(self):
        """Fire up the validation sequance"""
        self.controller.validate()

    def listen(self):
        """Listen on incoming messages from host

        TODO(marcus): We can't use this, as we are already listening on stdin
            through client.py. Do use this, we will have to find a way to
            receive multiple signals from the same stdin, and channel them
            to their corresponding source.

        """

        def _listen():
            while True:
                line = self.host.channels["parent"].get()
                payload = json.loads(line)["payload"]

                # We can't call methods directly, as we are running
                # in a thread. Instead, we emit signals that do the
                # job for us.
                signal = {

                    "show": "shown",
                    "hide": "hidden",
                    "quit": "quitted",
                    "publish": "published",
                    "validate": "validated",

                    "resize": "resized",

                    "rise": "risen",
                    "inFocus": "inFocused",
                    "outFocus": "outFocused",

                    "attach": "attached",
                    "detach": "detached",
                    "host_attach": "host_attached",
                    "host_detach": "host_detached",

                }.get(payload["name"])

                if not signal:
                    print("'{name}' was unavailable.".format(
                        **payload))
                else:
                    try:
                        getattr(self, signal).emit(
                            *payload.get("args", []))
                    except Exception:
                        traceback.print_exc()

        thread = threading.Thread(target=_listen)
        thread.daemon = True
        thread.start()


def main(demo=False, aschild=False, targets=[]):
    """Start the Qt-runtime and show the window

    Arguments:
        aschild (bool, optional): Run as child of parent process

    """

    if aschild:
        print("Starting pyblish-qml")
        compat.main()
        app = Application(APP_PATH, targets)
        app.listen()

        print("Done, don't forget to call `show()`")
        return app.exec_()

    else:
        print("Starting pyblish-qml server..")
        service = ipc.service.MockService() if demo else ipc.service.Service()
        server = ipc.server.Server(service, targets=targets)

        proxy = ipc.server.Proxy(server)
        proxy.show(settings.to_dict())

        server.listen()
        server.wait()
